package AOSI;

use strict;
use AOSICFG;
use Plugins;
AOSICFG::write_ws_log("-=- Back to Main -=-");
use Net::LDAP;
use Encode;
use XML::Simple;
use MIME::Base64;
use Digest::MD5 qw(md5 md5_base64 md5_hex);
use Digest::SHA qw(sha1_base64);

######################################################################
# Check if the service is alive
######################################################################
sub ldap_check_server {
    my ($service) = @_;
    my ($sock, $hostname, $port);

    ($hostname, $port) = AOSICFG::parse_server($service);
    use IO::Socket::INET;

    $sock = IO::Socket::INET->new(
        PeerPort => $port,
        PeerAddr => $hostname,
        Proto    => 'tcp',
        Timeout  => 60
    );
    return $sock;
}

sub t2t {
    my ($in) = @_;
    my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst, $out);

    ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = localtime($in);
    $out = sprintf("%04d-%02d-%02d %02d:%02d:%02d", $year + 1900, $mon + 1, $mday, $hour, $min, $sec);
    return $out;
}

######################################################################
# Get ldap num info
######################################################################
sub ldapInfo {
    my ($service, $user, $pwd, $base, $type) = @_;  # type: 0 - LDAP count, 1 - don't count
    my ($hash, $out_msg, $log, $dn, $ldap, $ldap_error, $dec_pwd, $mesg, $error, $code, $line);
    my ($filter, @entries, $port, $hostname, $num, $dpkg, $out, @out, $status, $name, $version, $desc);
    my($deb_ver, $deb_ver_file, $pkg);
    $log     = "ldapInfo";
    $out_msg = '';

    AOSICFG::write_ws_log("/ $log(in):user:$user;base:$base;type:$type;") if $config::aosi_debug_level > 0;
    AOSICFG::write_ws_log("| $log: IP:$AOSI::client_ip;")                 if $config::aosi_debug_level > 1;

    ($hostname, $port) = AOSICFG::parse_server($config::ldap_server);
    ($ldap, $ldap_error) = AOSICFG::ldap_init($hostname, $port);
    if ($user =~ /=/g) {
        $dn = $user . ',' . $base;
    }
    else {
        $dn = 'uid=' . $user . ',' . $base;
    }
    AOSICFG::write_ws_log("| $log: dn=$dn;") if $config::aosi_debug_level > 1;

    if (AOSICFG::mon_is_valid($AOSI::client_ip)) {
        if ($ldap) {
            $dec_pwd = decode_base64($pwd);
            $mesg    = $ldap->bind($dn, password => $dec_pwd, version => 3);
            $code    = $mesg->code;
            $error   = $mesg->error;
            if ($code == 0) {
                # Package info old
                #$ENV{'PATH'} = '/usr/bin:/bin';
                #$dpkg = 'dpkg -l \'*-aai\' |grep ^i';
                #$out  = qx($dpkg);
                #@out  = split(/[\n\r]/, $out);
                #AOSICFG::write_ws_log("| $log: run=$dpkg") if $config::aosi_debug_level > 3;
                #foreach $line (@out) {
                #    ($status, $name, $version, $desc) = split(/\s+/, $line);
                #    $out_msg .= '<attribute ldapname="package_'.$name.'"><value>' . $version . '</value></attribute>';
                #}
                #$deb_ver_file = '/etc/debian_version';
                #if(-e $deb_ver_file){
                #    $deb_ver = '/bin/cat '.$deb_ver_file;
                #    $deb_ver = qx($deb_ver);
                #}
                #else {
                #    $deb_ver = 'Other OS';
                #}
                #chomp($deb_ver);
                #$out_msg .= '<attribute ldapname="debian_version"><value>' . $deb_ver . '</value></attribute>';
                # Package info
                foreach $pkg (keys %AOSICFG::package_version){
                    $out_msg .= '<attribute ldapname="package_'.$pkg.'"><value>' . $AOSICFG::package_version{$pkg} . '</value></attribute>';
                }
                $out_msg .= '<attribute ldapname="debian_version"><value>' . $AOSICFG::debian_version . '</value></attribute>';
                if(($type == 0) or ($type eq '')){
                    $filter = '(uid=*)';
                    $mesg   = $ldap->search(base => $base, filter => $filter, timelimit => 60);
                    $code   = $mesg->code;
                    $error  = $mesg->error;
                    if ($code == 0) {
                        @entries = $mesg->entries;
                        $num     = $#entries;
                        AOSICFG::write_ws_log("| $log: num=$num;") if $config::aosi_debug_level > 2;
                        $out_msg .= '<attribute ldapname="num_of_uids"><value>' . $num . '</value></attribute>';
                    }
                    else {
                        $out_msg = '<status code="' . $code . '">' . $error . '</status>';
                    }
                }
                $mesg = $ldap->unbind;
                # AOSI config
                $out_msg .= '<attribute ldapname="aosi_debug_level"><value>' . $config::aosi_debug_level . '</value></attribute>';
                $out_msg .= '<attribute ldapname="direct_access"><value>' . $config::direct_access . '</value></attribute>';
            }
            else {
                $out_msg = '<status code="' . $code . '">' . $error . '</status>';
            }
        }
        else {
            $code    = -18;
            $error   = $ldap_error;
            $out_msg = '<status code="' . $code . '">' . $error . '</status>';
        }
    }
    else {
        $code    = -19;
        $error   = 'You are not allowed to call this function!';
        $out_msg = '<status code="' . $code . '">' . $error . '</status>';
    }
    $hash->{'code'}  = $code;
    $hash->{'error'} = $error;
    $hash->{'result'} =
      '<?xml version="1.0"?><ldap action="' . $log . '" version="' . $config::aosi_version . '"><entry dn="' . $dn . '">' . $out_msg . '</entry></ldap>';

    #AOSICFG::write_action_log($user, $log, '', $code, $base);
    AOSICFG::write_ws_log("\\ $log(out-mod):" . AOSICFG::dump_hash($hash)) if $config::aosi_debug_level > 0;
    return hash2data($hash, { 'code' => 'int' });
}

######################################################################
# Get admin info
######################################################################
sub ldapAdminInfo {
    my ($service, $user, $pwd, $base, $admin_uid, $type) = @_;
    my ($hash,    $out_msg, $log,     $uid,     $attrib,  $ldap,      $mesg,       $error,    $code);
    my (@entries, $entry,   @values,  $filter,  $dec_pwd, $hostname,  $port,       $dummy,    $dn);
    my ($val,     @urn,     $urn_key, $urn_val, $cur_val, $auth_code, $auth_error, $is_admin, $ldap_error);
    $log = "ldapAdminInfo";

    AOSICFG::write_ws_log("/ $log(in):user:$user;base:$base;admin_uid:$admin_uid;type:$type;") if $config::aosi_debug_level > 0;
    AOSICFG::write_ws_log("| $log: IP:$AOSI::client_ip;")                                      if $config::aosi_debug_level > 1;

    ($hostname, $port) = AOSICFG::parse_server($config::ldap_server);
    ($ldap, $ldap_error) = AOSICFG::ldap_init($hostname, $port);
    $dn = AOSICFG::get_dn($user, $base);
    ($auth_code, $auth_error, $is_admin) = AOSICFG::ck_auth($ldap, AOSICFG::get_dn($user, $base), $pwd, 1, $ldap_error);
    if ($is_admin == 0 and $ldap) {
        if ($auth_code == 0) {
            $auth_code  = -15;
            $auth_error = "User $user does not have admin privileges!";
        }
    }
    if ($auth_code == 0) {

        ####################### Valid user file #########################
        AOSICFG::write_ws_log("| $log(valid_users): " . join(",", @AOSICFG::ldap_admins)) if $config::aosi_debug_level > 1;
        foreach $uid (@AOSICFG::ldap_admins) {
            next if ($uid eq '');
            if ($admin_uid eq '') {
                $out_msg .= '<entry dn="uid=' . $uid . ',' . $base . '" admintype="1">';
                $out_msg .= '</entry>';
            }
            else {
                if ($uid eq $admin_uid) {
                    $out_msg .= '<entry dn="uid=' . $uid . ',' . $base . '" admintype="1">';
                    $out_msg .= '</entry>';
                }
            }
        }
        AOSICFG::write_ws_log("| $log(out_msg valid_user): $out_msg") if $config::aosi_debug_level > 1;

        ####################### LDAP: hrEduPersonRole=administrator imenika #########################
        $dec_pwd = decode_base64($pwd);
        $mesg    = $ldap->bind(AOSICFG::get_dn($user, $base), password => $dec_pwd, version => 3);
        $code    = $mesg->code;
        $error   = $mesg->error;
        $attrib  = 'hrEduPersonRole';
        $filter  = '(' . $attrib . '=administrator imenika)';
        $mesg    = $ldap->search(base => $base, filter => $filter, timelimit => 60, attrs => [$attrib]);
        $code    = $mesg->code;
        $error   = $mesg->error;
        @entries = $mesg->entries;
        AOSICFG::write_ws_log("| $log(entries hrEduPersonRole=administrator imenika): " . $#entries) if $config::aosi_debug_level > 1;

        foreach $entry (@entries) {
            $dn = $entry->dn();
            ($uid, $dummy) = AOSICFG::get_uid_base($dn);
            AOSICFG::write_ws_log("| $log(hrEduPersonRole=administrator imenika): dn:$dn;uid:$uid;") if $config::aosi_debug_level > 2;
            if ($admin_uid eq '') {
                $out_msg .= '<entry dn="' . $dn . '" admintype="2">';
                $out_msg .= '</entry>';
            }
            else {
                if ($uid eq $admin_uid) {
                    $out_msg .= '<entry dn="' . $dn . '" admintype="2">';
                    $out_msg .= '</entry>';
                }
            }
        }
        AOSICFG::write_ws_log("| $log(out_msg hrEduPersonRole=administrator imenika): $out_msg") if $config::aosi_debug_level > 1;

        ####################### LDAP: hrEduPersonRole=urn:* #########################
        $attrib  = 'hrEduPersonRole';
        $filter  = '(' . $attrib . '=urn:administrator imenika*)';
        $mesg    = $ldap->search(base => $base, filter => $filter, timelimit => 60, attrs => [$attrib]);
        $code    = $mesg->code;
        $error   = $mesg->error;
        @entries = $mesg->entries;
        AOSICFG::write_ws_log("| $log(entries hrEduPersonRole=urn:*): " . $#entries) if $config::aosi_debug_level > 1;
        foreach $entry (@entries) {
            $dn = $entry->dn();
            ($uid, $dummy) = AOSICFG::get_uid_base($dn);
            AOSICFG::write_ws_log("| $log(hrEduPersonRole=urn:*): dn:$dn;uid:$uid;") if $config::aosi_debug_level > 2;
            @values = @{ $entry->get_value($attrib, asref => 1) };
            AOSICFG::write_ws_log("| $log(hrEduPersonRole=urn:*): values:" . join(",", @values)) if $config::aosi_debug_level > 2;
            if ($admin_uid eq '') {
                $out_msg .= '<entry dn="' . $dn . '" admintype="3">';
                foreach $val (@values) {
                    if ($val =~ /^urn\:*/) {
                        $out_msg .= '<urn>';
                        @urn = split(/:/, $val);
                        AOSICFG::write_ws_log("| $log(hrEduPersonRole=urn:*): urn:" . join(",", @urn)) if $config::aosi_debug_level > 2;
                        $dummy = shift(@urn);
                        $dummy = shift(@urn);
                        if ($#urn > -1) {
                            while ($urn_key = shift(@urn)) {
                                $urn_val = shift(@urn);
                                $out_msg .= '<urn_value ldapname="' . $urn_key . '">' . $urn_val . '</urn_value>';
                            }
                        }
                        else {
                            $out_msg .= '<urn_value ldapname="administrator imenika"></urn_value>';
                        }
                        $out_msg .= '</urn>';
                    }
                }
                $out_msg .= '</entry>';
            }
            else {
                if ($uid eq $admin_uid) {
                    $out_msg .= '<entry dn="' . $dn . '" admintype="3">';
                    foreach $val (@values) {
                        if ($val =~ /^urn\:*/) {
                            $out_msg .= '<urn>';
                            @urn = split(/:/, $val);
                            AOSICFG::write_ws_log("| $log(hrEduPersonRole=urn:*): urn:" . join(",", @urn)) if $config::aosi_debug_level > 2;
                            $dummy = shift(@urn);
                            $dummy = shift(@urn);
                            if ($#urn > -1) {
                                while ($urn_key = shift(@urn)) {
                                    $urn_val = shift(@urn);
                                    $out_msg .= '<urn_value ldapname="' . $urn_key . '">' . $urn_val . '</urn_value>';
                                }
                            }
                            else {
                                $out_msg .= '<urn_value ldapname="administrator imenika"></urn_value>';
                            }
                            $out_msg .= '</urn>';
                        }
                    }
                    $out_msg .= '</entry>';
                }
            }
        }
        AOSICFG::write_ws_log("| $log(out_msg hrEduPersonRole=urn*): $out_msg") if $config::aosi_debug_level > 1;
        if ($out_msg eq '') {
            $out_msg .= '<entry dn="uid=' . $admin_uid . ',' . $base . '" admintype="0"></entry>';
        }

        $hash->{'code'}   = $code;
        $hash->{'error'}  = '0 error(s) occured!';
        $hash->{'result'} = '<ldap action="' . $log . '" version="' . $config::aosi_version . '">' . $out_msg . '</ldap>';
        $ldap->unbind;
    }
    else {
        $hash->{'code'}  = $auth_code;
        $hash->{'error'} = $auth_error;
        $hash->{'result'} =
            '<?xml version="1.0"?><ldap action="' . $log
          . '" version="'
          . $config::aosi_version
          . '"><entry dn="'
          . $dn
          . '"><status code="'
          . $auth_code . '">'
          . $auth_error
          . '</status></entry></ldap>';
    }

    AOSICFG::write_action_log($user, $log, $admin_uid, $code, $base);
    if ($type > 0) {
        AOSICFG::write_ws_log("\\ $log(out):" . AOSICFG::dump_hash($hash)) if $config::aosi_debug_level > 0;
        return $hash;
    }
    else {
        AOSICFG::write_ws_log("\\ $log(out-mod):" . AOSICFG::dump_hash($hash)) if $config::aosi_debug_level > 0;
        return hash2data($hash, { 'code' => 'int' });
    }
}

######################################################################
# Add hrEduOrg attributes
######################################################################
sub ldapOrgAdd {
    my ($service, $user, $pwd, $base, $xml_in) = @_;
    my ($hash, $log);
    $log = "ldapOrgAdd";

    AOSICFG::write_ws_log("/ $log(in):user:$user;base:$base;xml:$xml_in;") if $config::aosi_debug_level > 0;
    AOSICFG::write_ws_log("| $log: IP:$AOSI::client_ip;")                  if $config::aosi_debug_level > 1;
    $hash = AOSICFG::ldapOrgGenModify($log, $user, $pwd, $base, $xml_in, 1);
    AOSICFG::write_ws_log("\\ $log(out-mod):" . AOSICFG::dump_hash($hash)) if $config::aosi_debug_level > 0;
    return hash2data($hash, { 'code' => 'int' });
}

######################################################################
# Modify hrEduOrg attributes
######################################################################
sub ldapOrgModify {
    my ($service, $user, $pwd, $base, $xml_in) = @_;
    my ($hash, $log);
    $log = "ldapOrgModify";

    AOSICFG::write_ws_log("/ $log(in):user:$user;base:$base;xml:$xml_in;") if $config::aosi_debug_level > 0;
    AOSICFG::write_ws_log("| $log: IP:$AOSI::client_ip;")                  if $config::aosi_debug_level > 1;
    $hash = AOSICFG::ldapOrgGenModify($log, $user, $pwd, $base, $xml_in, 2);
    AOSICFG::write_ws_log("\\ $log(out-mod):" . AOSICFG::dump_hash($hash)) if $config::aosi_debug_level > 0;
    return hash2data($hash, { 'code' => 'int' });
}

######################################################################
# Delete hrEduOrg attributes
######################################################################
sub ldapOrgDelete {
    my ($service, $user, $pwd, $base, $xml_in) = @_;
    my ($hash, $log);
    $log = "ldapOrgDelete";

    AOSICFG::write_ws_log("/ $log(in):user:$user;base:$base;xml:$xml_in;") if $config::aosi_debug_level > 0;
    AOSICFG::write_ws_log("| $log: IP:$AOSI::client_ip;")                  if $config::aosi_debug_level > 1;
    $hash = AOSICFG::ldapOrgGenModify($log, $user, $pwd, $base, $xml_in, 3);
    AOSICFG::write_ws_log("\\ $log(out-mod):" . AOSICFG::dump_hash($hash)) if $config::aosi_debug_level > 0;
    return hash2data($hash, { 'code' => 'int' });
}

######################################################################
# Get hrEduOrg info, i.e. following attributes o,l,postalAddress
######################################################################
sub ldapOrgInfo {
    my ($service, $user, $pwd, $base, $type) = @_;
    my ($hash, $log, $filter, $attributes);
    $log = "ldapOrgInfo";

    AOSICFG::write_ws_log("/ $log(in):user:$user;base:$base;type:$type;") if $config::aosi_debug_level > 0;
    AOSICFG::write_ws_log("| $log: IP:$AOSI::client_ip;")                 if $config::aosi_debug_level > 1;

    $filter     = '(objectclass=organization)';
    $attributes = '*';
    $hash       = AOSICFG::ldapGenSearch($log, $user, $pwd, $base, $filter, $attributes, 0, -1, 0);
    if ($type > 0) {
        AOSICFG::write_ws_log("\\ $log(out):" . AOSICFG::dump_hash($hash)) if $config::aosi_debug_level > 0;
        return $hash;
    }
    else {
        AOSICFG::write_ws_log("\\ $log(out-mod):" . AOSICFG::dump_hash($hash)) if $config::aosi_debug_level > 0;
        return hash2data($hash, { 'code' => 'int' });
    }
}

######################################################################
# Get hrEduOrg info, i.e. following attributes o,l,postalAddress
######################################################################
sub userOrgInfo {
    my ($service, $user, $pwd, $base, $type) = @_;
    my ($hash, $log, $filter, $attributes);
    $log = "userOrgInfo";

    AOSICFG::write_ws_log("/ $log(in):user:$user;base:$base;filter:$filter;attrib:$attributes;") if $config::aosi_debug_level > 0;
    AOSICFG::write_ws_log("| $log: IP:$AOSI::client_ip;")                                        if $config::aosi_debug_level > 1;

    $filter     = '(objectclass=organization)';
    $attributes = '*';
    $hash       = AOSICFG::ldapGenSearch($log, $user, $pwd, $base, $filter, $attributes, 1, -1, 0);
    if ($type > 0) {
        AOSICFG::write_ws_log("\\ $log(out):" . AOSICFG::dump_hash($hash)) if $config::aosi_debug_level > 0;
        return $hash;
    }
    else {
        AOSICFG::write_ws_log("\\ $log(out-mod):" . AOSICFG::dump_hash($hash)) if $config::aosi_debug_level > 0;
        return hash2data($hash, { 'code' => 'int' });
    }
}

######################################################################
# Just bind to LDAP
######################################################################
sub ldapBind {
    my ($service, $user, $pwd, $base, $type) = @_;
    my ($hostname, $port, $ldap, $mesg, $code, $hash, $mesg2, $log, $out);
    my ($auth_code, $auth_error, $is_admin_ldap, $ldap_error);
    $log = "ldapBind";
    if ($service eq 'userBind') {
        $log = $service . '::' . $log;
    }

    AOSICFG::write_ws_log("/ $log(in):user:$user;base:$base");

    ($hostname, $port) = AOSICFG::parse_server($config::ldap_server);
    ($ldap, $ldap_error) = AOSICFG::ldap_init($hostname, $port);

    ($auth_code, $auth_error, $is_admin_ldap) = AOSICFG::ck_auth($ldap, AOSICFG::get_dn($user, $base), $pwd, 1, $ldap_error);

    #$admin = AOSICFG::user_is_valid($user);
    AOSICFG::write_ws_log("| $log(admin type): $is_admin_ldap");
    $hash->{'code'}   = $auth_code;
    $hash->{'error'}  = $auth_error;
    $hash->{'result'} = $is_admin_ldap;
    if ($type > 0) {
        AOSICFG::write_ws_log("\\ $log(out):" . AOSICFG::dump_hash($hash));
        return $hash;
    }
    else {
        AOSICFG::write_ws_log("\\ $log(out-mod):" . AOSICFG::dump_hash($hash));
        return hash2data($hash, { 'code' => 'int', 'result' => 'int' });
    }
}

######################################################################
# Just bind to LDAP
######################################################################
sub userBind {
    my ($service, $user, $pwd, $base) = @_;
    my ($code, $hash, $log, $bind_hash);
    $log = "userBind";

    AOSICFG::write_ws_log("/ $log(in):user:$user;base:$base");
    $hash = ldapBind($log, $user, $pwd, $base, 1);
    AOSICFG::write_ws_log("\\ $log(out):" . AOSICFG::dump_hash($hash));
    return hash2data($hash, { 'code' => 'int', 'result' => 'int' });
}

######################################################################
# Delete user(s) from LDAP
######################################################################
sub ldapDeleteUser {
    my ($service, $user, $pwd, $base, $xml_in) = @_;
    my ($attr_exists, $auth_code, $auth_error, $code, $cur_code, $dn, $ent, $entries);
    my ($err_num, $hash, $hostname, $is_admin, $ldap, $log, $mesg2, $msg, $out_msg, $port, $admin_filter);
    my ($uid,        $my_filter,   @entries2,     $ldap_error, $out_one);
    my ($plugin_key, $plugin_hash, $plugin_panic, $plugin_xml, $plugin_code);
    $log     = "ldapDeleteUser";
    $out_msg = '';
    $code    = 0;

    AOSICFG::write_ws_log("/ $log(in):user:$user;base:$base;xml:$xml_in") if $config::aosi_debug_level > 0;
    AOSICFG::write_ws_log("| $log: IP:$AOSI::client_ip;")                 if $config::aosi_debug_level > 1;

    ($hostname, $port) = AOSICFG::parse_server($config::ldap_server);
    ($ldap, $ldap_error) = AOSICFG::ldap_init($hostname, $port);
    $dn = AOSICFG::get_dn($user, $base);
    ($auth_code, $auth_error, $is_admin, $admin_filter) = AOSICFG::ck_auth($ldap, AOSICFG::get_dn($user, $base), $pwd, 1, $ldap_error);
    Encode::_utf8_on($admin_filter);
    if ($is_admin == 0 and $ldap) {
        if ($auth_code == 0) {
            $auth_code  = -15;
            $auth_error = "User $user does not have admin privileges!";
        }
    }
    if ($auth_code == 0) {
        ($entries, $attr_exists) = AOSICFG::xml2entries($xml_in);
        $err_num = 0;
        foreach $ent (@$entries) {
            $dn = AOSICFG::utfize($ent->{'dn'});
            if ($dn) {
                ($uid, $base) = AOSICFG::get_uid_base($dn);
                $uid = AOSICFG::utfize($uid);
                $my_filter = '(&(uid=' . $uid . ')' . $admin_filter . ')';
                AOSICFG::write_ws_log("| $log:my_filter=$my_filter;") if $config::aosi_debug_level > 2;
                $mesg2 = $ldap->bind($config::ldap_user, password => AOSICFG::get_pwd($config::ldap_pwd_file), version => 3);
                $cur_code = $mesg2->code;
                $out_one = '<entry dn="' . $dn . '">';
                if ($cur_code == 0) {
                    $mesg2    = $ldap->search(base => $base, filter => $my_filter, timelimit => 60, attrs => ['uid']);
                    $cur_code = $mesg2->code;
                    @entries2 = $mesg2->entries;
                    AOSICFG::write_ws_log("| $log(search with my_filter):code:$cur_code;error:" . $mesg2->error . ";entries:" . $#entries2)
                      if $config::aosi_debug_level > 2;
                    if ($cur_code == 0) {
                        ($plugin_hash) = Plugins::beforeDeleteUser($ldap, $base, $ent, $user, $pwd);
                        ($plugin_panic, $plugin_code, $plugin_xml) = AOSICFG::process_plugin_hash($plugin_hash, 'before');
                        $out_one .= $plugin_xml;
                        if ($plugin_panic == 0) {
                            if ($#entries2 > -1) {
                                $mesg2    = $ldap->delete($dn);
                                $cur_code = $mesg2->code;
                                $code += $cur_code;
                                $msg = $mesg2->error;
                                if ($cur_code == 0) {
                                    $out_one .= '<status code="' . $cur_code . '">User deleted successfully!</status>';
                                }
                                else {
                                    $err_num++;
                                    $out_one .= '<status code="' . $cur_code . '">' . $msg . '!</status>';
                                }
                            }
                            else {
                                $code += 81;
                                $cur_code = -81;
                                $out_one .= '<status code="-81">User does not exist!</status>';
                                $err_num++;
                            }
                            ($plugin_hash) = Plugins::afterDeleteUser($ent, $out_one . '</entry>', $user, $pwd);
                            ($plugin_panic, $plugin_code, $plugin_xml) = AOSICFG::process_plugin_hash($plugin_hash, 'after');
                            $out_one .= $plugin_xml;
                        }
                        else {
                            $code += abs($plugin_code);
                            $out_one .= '<status code="' . $code . '">Plugin stopped the action!</status>';
                            $err_num++;
                        }
                    }
                    else {
                        $code += $cur_code;
                        $out_one .= '<status code="' . $cur_code . '">' . $mesg2->error . '!</status>';
                        $err_num++;
                    }
                }
                else {
                    $code += $cur_code;
                    $out_one .= '<status code="' . $cur_code . '">' . $mesg2->error . '!</status>';
                    $err_num++;
                }
            }
            $out_one .= '</entry>';
            $out_msg .= $out_one;
            AOSICFG::write_action_log($user, $log, $dn, $cur_code);
        }
        if ($code == 0) {
            $hash->{'code'}  = $code;
            $hash->{'error'} = '';
        }
        else {
            $hash->{'code'}  = $code;
            $hash->{'error'} = $err_num . ' error(s) occured!';
        }
        $hash->{'result'} = '<?xml version="1.0"?><ldap action="' . $log . '" version="' . $config::aosi_version . '">' . $out_msg . '</ldap>';
    }
    else {
        $hash->{'code'}  = $auth_code;
        $hash->{'error'} = $auth_error;
        $hash->{'result'} =
            '<?xml version="1.0"?><ldap action="' . $log
          . '" version="'
          . $config::aosi_version
          . '"><entry dn="'
          . $dn
          . '"><status code="'
          . $auth_code . '">'
          . $auth_error
          . '</status></entry></ldap>';
    }

    AOSICFG::write_ws_log("\\ $log(out):" . AOSICFG::dump_hash($hash)) if $config::aosi_debug_level > 0;

    #return $hash;
    return hash2data($hash, { 'code' => 'int' });
}

######################################################################
# Add user to LDAP Light Edition
# Attributes l, o, postalAddress, hrEduPersonUniqueID, hrEduPersonHomeOrg are autofilled
# Attributes are in XML (http://www.aaiedu.hr/aosi/ldap.dtd)
######################################################################
sub ldapAddUserLE {
    my ($service, $user, $pwd, $base, $xml_in) = @_;
    my ($att_hash, $attr,    $attr_exists, $code, $dc,  $dn,      $domain,          $ent,         $ent_attrs);
    my ($entries,  $err_num, $error,       $hash, $key, $log,     $org_attr_exists, $org_entries, $org_hash, $out);
    my ($result,   $tld,     $uid,         $val,  $xml, %org_att, %removed_attr,    @no_attr,     @remove_attr);
    my (@req_attr, @req_attrs, $hep_PA, $i, $OIB, $tmp_code, $tmp_msg, $cn, $disp_name, $hep_un, $OIB_un);
    $log  = "ldapAddUserLE";
    $code = 0;
    $OIB = '';

    ($domain, $tld)    = split(/,/, $base);
    ($dc,     $domain) = split(/=/, $domain);
    ($dc,     $tld)    = split(/=/, $tld);
    $domain = $domain . '.' . $tld;

    @remove_attr = ('o', 'l', 'postalAddress', 'hrEduPersonUniqueID', 'hrEduPersonHomeOrg');
    foreach $attr (@remove_attr) {
        $removed_attr{$attr}++;
    }
    foreach $attr (@AOSICFG::req_attribs) {
        if ($removed_attr{$attr} < 1) {
            push(@req_attr, $attr);
        }
    }

    AOSICFG::write_ws_log("/ $log(in):user:$user;base:$base;xml:$xml_in") if $config::aosi_debug_level > 0;
    AOSICFG::write_ws_log("| $log: IP:$AOSI::client_ip;")                 if $config::aosi_debug_level > 1;

    ($entries, $attr_exists) = AOSICFG::xml2entries($xml_in);

    #$org_hash = ldapOrgInfo($service, $user, $pwd, $base, 1);
    $org_hash = userOrgInfo($service, $user, $pwd, $base, 1);
    $code     = $org_hash->{'code'};
    $result   = $org_hash->{'result'};
    $error    = $org_hash->{'error'};
    $err_num  = 0;
    if ($code == 0) {
        ($org_entries, $org_attr_exists) = AOSICFG::xml2entries($result);
        $ent_attrs = $org_entries->[0]->{'attrs'};
        foreach $att_hash (@$ent_attrs) {
            $key           = $att_hash->{'att'};
            $val           = $att_hash->{'val'};
            $org_att{$key} = $val;
        }

        foreach $ent (@$entries) {
            $attr_exists = undef;
            $dn          = AOSICFG::utfize($ent->{'dn'});
            $ent_attrs   = $ent->{'attrs'};
            foreach $key (@$ent_attrs) {
                if ($key->{'att'} eq 'uid') {
                    $uid = $key->{'val'}->[0];
                }
                if ($key->{'att'} eq 'cn') {
                    $cn = $key->{'val'}->[0];
                }
                if ($key->{'att'} eq 'hrEduPersonOIB') {
                    $OIB = $key->{'val'}->[0];
                }
                if ($key->{'att'} eq 'hrEduPersonUniqueNumber') {
                    foreach my $hep_un (@{$key->{'val'}}){
                        if($hep_un =~ /^OIB:\s*(\d+)/){
                            $OIB_un = $1;
                            AOSICFG::write_ws_log("| $log(OIB detected in hEPUN): hep_un=$hep_un;OIB_un=$OIB_un;") if $config::aosi_debug_level > 3;
                        }
                    }
                }
                if ($key->{'att'} eq 'hrEduPersonPrimaryAffiliation') {
                    $hep_PA = $key->{'val'}->[0];
                }
            }
            foreach $attr (@$ent_attrs) {
                $attr_exists->{ $attr->{'att'} }++;
            }
            @no_attr = ();
            foreach $attr (@req_attrs) {
                if ($attr_exists->{$attr} < 1) {
                    push(@no_attr, $attr);
                }
            }
            if ($#no_attr > -1) {
                $out =
                    '<?xml version="1.0"?><ldap action="' . $log
                  . '" version="'
                  . $config::aosi_version
                  . '"><entry dn="'
                  . $dn
                  . '"><status code="-25">Missing attributes '
                  . join(", ", @no_attr)
                  . '!</status></entry></ldap>';
                $code += 25;
                $err_num++;
                $hash->{'result'} .= $out;
                AOSICFG::write_ws_log("| $log(out):" . AOSICFG::dump_hash($hash)) if $config::aosi_debug_level > 0;
                AOSICFG::write_ws_log("\\ $log(error): Missing attributes " . join(", ", @no_attr) . "!");
            }
            else {
                $ent_attrs = $ent->{'attrs'};
                @no_attr   = ();
                foreach $i (0 .. $#{$ent_attrs}) {
                    $attr = $ent_attrs->[$i];
                    if (   ($attr->{'att'} eq 'o')
                        or ($attr->{'att'} eq 'l')
                        or ($attr->{'att'} eq 'postalAddress')
                        or ($attr->{'att'} eq 'hrEduPersonHomeOrg')
                        or ($attr->{'att'} eq 'hrEduPersonAffiliation')
                        or ($attr->{'att'} eq 'hrEduPersonUniqueID'))
                    {
                        push(@no_attr, $i);
                        AOSICFG::write_ws_log("| $log(removing $i):" . $attr->{'att'}) if $config::aosi_debug_level > 3;
                    }
                }
                foreach $i (reverse 0 .. $#no_attr) {    # Delete in reverse because splice shortens the array, this way indexes do not shift
                    splice(@$ent_attrs, $no_attr[$i], 1);
                }
                push(@$ent_attrs, { 'att' => 'o',                      'val' => $org_att{'o'} });
                push(@$ent_attrs, { 'att' => 'l',                      'val' => $org_att{'l'} });
                push(@$ent_attrs, { 'att' => 'postalAddress',          'val' => $org_att{'postalAddress'} });
                push(@$ent_attrs, { 'att' => 'hrEduPersonHomeOrg',     'val' => [$domain] });
                push(@$ent_attrs, { 'att' => 'hrEduPersonAffiliation', 'val' => [$hep_PA] });
                push(@$ent_attrs, { 'att' => 'hrEduPersonUniqueID',    'val' => [$uid . '@' . $domain] });
                if(AOSICFG::aosi_version(2) > 2){ # AOSI version 3 is hrEduPerson 1.3 compatible
                    push(@$ent_attrs, { 'att' => 'hrEduPersonPersistentID','val' => [md5_hex($uid . '@' . $domain.time)] });
                    $disp_name = $cn;
                    # tr does not work in tainted mode
                    # so this is substituted
                    #$disp_name =~ tr/\x{0106}\x{0107}\x{010C}\x{010D}\x{0110}\x{0111}\x{0160}\x{0161}\x{017D}\x{017E}/CcCcDdSsZz/;
                    $disp_name =~ s/\x{0106}/C/g;
                    $disp_name =~ s/\x{0107}/c/g;
                    $disp_name =~ s/\x{010C}/C/g;
                    $disp_name =~ s/\x{010D}/c/g;
                    $disp_name =~ s/\x{0110}/D/g;
                    $disp_name =~ s/\x{0111}/d/g;
                    $disp_name =~ s/\x{0160}/S/g;
                    $disp_name =~ s/\x{0161}/s/g;
                    $disp_name =~ s/\x{017D}/Z/g;
                    $disp_name =~ s/\x{017E}/z/g;
                    push(@$ent_attrs, { 'att' => 'displayName','val' => [$disp_name] });
                    foreach $key (@$ent_attrs) {
                        if ($key->{'att'} eq 'hrEduPersonOIB') {
                            # Add OIB - ldapAddUser will take care of restrictions
                            push(@$ent_attrs, { 'att' => 'hrEduPersonUniqueNumber','val' => ['OIB: '.$OIB] });
                        }
                    }
                    if(($OIB eq '') and ($OIB_un ne '')){
                        push(@$ent_attrs, { 'att' => 'hrEduPersonOIB',    'val' => [$OIB_un] });
                    }
                }
            }
        }
        if ($code == 0) {
            AOSICFG::write_ws_log("| $log(sending to AddUser):" . AOSICFG::dump_hash($entries->[0])) if $config::aosi_debug_level > 0;
            $hash = ldapAddUser($log, $user, $pwd, $base, $entries, 1, 1);
        }
        else {
            $hash->{'code'}  = $code;
            $hash->{'error'} = $err_num . ' error(s) occured!';
            $hash->{'result'} = $result;
        }
    }
    else {
        $hash->{'code'}   = $code;
        $hash->{'error'}  = $error;
        $hash->{'result'} = $result;
    }

    AOSICFG::write_ws_log("\\ $log(out):" . AOSICFG::dump_hash($hash)) if $config::aosi_debug_level > 0;

    #return $hash;
    return hash2data($hash, { 'code' => 'int' });
}

######################################################################
# Add user to LDAP
# Attributes are in XML (http://www.aaiedu.hr/aosi/ldap.dtd)
######################################################################
sub ldapAddUser {
    my ($service, $user, $pwd, $base, $xml_in, $type, $use_entries) = @_;
    my ($attr,         $attr_exists, $attrs,     $auth_code, $auth_error, $code,    $dn);
    my ($dn_in,        $ent,         $ent_attrs, $entries,   $err_num,    $hash,    $hostname, $is_admin);
    my ($ldap,         $log,         $mesg2,     $out,       $port,       $val,     @no_attr, @req_attrs);
    my ($admin_filter, $cond,        $key,       $urn_key,   $urn_val,    $user_ok, $user_ok2, $user_ok3, @val);
    my ($att_name, $att_val, $i, $j, $att_arr, $urn_val_no, $ldap_error, $out_one, $tmp_msg, $tmp_code);
    my ($plugin_key, $plugin_hash, $plugin_panic, $plugin_xml, $plugin_code, $restriction_code, $restriction_msg);
    $log       = $service == 'ldapAddUserLE' ? "ldapAddUser::ldapAddUserLE" : "ldapAddUser";
    $code      = 0;
    @req_attrs = @AOSICFG::req_attribs;

    AOSICFG::write_ws_log("/ $log(in):user:$user;base:$base;dn:$dn_in;xml:$xml_in") if $config::aosi_debug_level > 0;
    AOSICFG::write_ws_log("| $log: IP:$AOSI::client_ip;")                           if $config::aosi_debug_level > 1;

    if ($use_entries == 1) {
        $entries = $xml_in;
        #foreach $key (@{$entries->[0]->{'attrs'}}){
        #    AOSICFG::write_ws_log("| $log(entries raw): " . AOSICFG::dump_hash($key)) if $config::aosi_debug_level > 3;
        #}
    }
    else {

        # Parsing XML
        ($entries, $attr_exists) = AOSICFG::xml2entries($xml_in);
    }

    ($hostname, $port) = AOSICFG::parse_server($config::ldap_server);
    ($ldap, $ldap_error) = AOSICFG::ldap_init($hostname, $port);
    $dn = AOSICFG::get_dn($user, $base);
    ($auth_code, $auth_error, $is_admin, $admin_filter, $cond) = AOSICFG::ck_auth($ldap, AOSICFG::get_dn($user, $base), $pwd, 1, $ldap_error);
    Encode::_utf8_on($admin_filter);
    if ($is_admin == 0 and $ldap) {
        if ($auth_code == 0) {
            $auth_code  = -15;
            $auth_error = "User $user does not have admin privileges!";
        }
    }
    if ($auth_code == 0) {
        $mesg2 = $ldap->bind($config::ldap_user, password => AOSICFG::get_pwd($config::ldap_pwd_file), version => 3);
        if ($mesg2->code != 0) {
            AOSICFG::write_ws_log("| !!!$log: invalid admin password in " . $config::ldap_pwd_file);
        }
        $hash->{'result'} = '<?xml version="1.0"?><ldap action="' . $service . '" version="' . $config::aosi_version . '">';
        $err_num = 0;
        foreach $ent (@$entries) {
            $dn      = AOSICFG::utfize($ent->{'dn'});
            $out_one = '<entry dn="' . $dn . '">';
            ($plugin_hash) = Plugins::beforeAddUser($ldap, $base, $ent, $user, $pwd);
            ($plugin_panic, $plugin_code, $plugin_xml) = AOSICFG::process_plugin_hash($plugin_hash, 'before');
            $out_one .= $plugin_xml;
            if ($plugin_panic == 0) {
                $attr_exists = undef;
                $ent_attrs   = $ent->{'attrs'};
                foreach $attr (@$ent_attrs) {
                    $attr_exists->{ $attr->{'att'} }++;
                }
                @no_attr = ();
                foreach $attr (@req_attrs) {
                    if ($attr_exists->{$attr} < 1) {
                        push(@no_attr, $attr);
                    }
                }
                if ($#no_attr > -1) {
                    $out_one .= '<status code="-22">Missing attributes ' . join(", ", @no_attr) . '!</status>';
                    $code += 22;
                    $err_num++;
                    AOSICFG::write_ws_log("| $log(out):" . AOSICFG::dump_hash($hash)) if $config::aosi_debug_level > 0;
                    AOSICFG::write_ws_log("\\ $log(error): Missing attributes " . join(", ", @no_attr) . "!") if $config::aosi_debug_level > 0;
                }
                else {
                    $attrs            = [];
                    $restriction_code = 0;
                    $restriction_msg  = '';
                    $j                = 0;
                    foreach $attr (@$ent_attrs) {
                        $val      = $attr->{'val'};
                        $att_name = $attr->{'att'};
                        if (AOSICFG::is_bin($att_name)) {
                            $val = decode_base64($val);
                        }
                        if ($att_name eq 'userPassword') {
                            $val = [ AOSICFG::encode_pwd($val->[0]) ];
                            $ent->{'attrs'}->[$j]->{'val'} = $val;
                        }
                        push(@$attrs, $att_name => $val);
                        if (ref($val) eq 'ARRAY') {
                            foreach $att_val (@$val) {
                                ($tmp_code, $tmp_msg) = AOSICFG::is_restriction_ok($att_name, $att_val);
                                if ($tmp_code == 0) {
                                    $restriction_code++;
                                    $restriction_msg .= ' ['.$att_name.'] '.$tmp_msg;
                                }
                            }
                        }
                        else {
                            ($tmp_code, $tmp_msg) = AOSICFG::is_restriction_ok($att_name, $val);
                            if ($tmp_code == 0) {
                                $restriction_code++;
                                $restriction_msg .= ' ['.$att_name.'] '.$tmp_msg;
                            }
                        }
                        $j++;
                    }
                    if ($restriction_code == 0) {
                        push(@$attrs, 'objectClass' => ['hrEduPerson', 'schacContactLocation']);
                        if ($is_admin == 1 or $is_admin == 2) {
                            AOSICFG::write_ws_log("| $log: user is admin: " . $is_admin) if $config::aosi_debug_level > 2;
                            $user_ok = 1;
                        }
                        else {

                            # Here we are checking if the admin can add a user
                            $user_ok = 0;

                            #AOSICFG::write_ws_log("| $log(check urn): ".AOSICFG::dump_hash($cond->[0])) if $config::aosi_debug_level > 2;
                            # Go through each urn:administrator:*
                            foreach $key (@$cond) {
                                $user_ok2 = 0;
                                # Chack every attribute that is in urn:*
                                foreach $urn_key (keys %$key) {
                                    $urn_val_no = $#{ $key->{$urn_key} };
                                    $user_ok3   = 0;
                                    # Go through all attributes that should be added
                                    foreach $i (0 .. $#{$attrs}) {

                                        # Odd are attribute names
                                        if ($i % 2 == 0) {
                                            $att_name = $attrs->[$i];
                                        }

                                        # Even are attribute values
                                        else {
                                            $att_arr = $attrs->[$i];
                                            if ($urn_key eq $att_name) {
                                                foreach $j (0 .. $urn_val_no) {
                                                    $urn_val = $key->{$urn_key}->[$j];
                                                    Encode::_utf8_on($urn_val);
                                                    foreach $att_val (@$att_arr) {
                                                        if ($att_val eq $urn_val) {
                                                            $user_ok3++;
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                    }
                                    AOSICFG::write_ws_log("| $log(cond 3):user_ok3=$user_ok3;") if $config::aosi_debug_level > 3;

                                    # Any of values is matched so this attribute is OK
                                    if ($user_ok3 > 0) {
                                        $user_ok2++;
                                    }
                                    AOSICFG::write_ws_log("| $log(cond finished att):  user_ok2:$user_ok2;user_ok3:$user_ok3") if $config::aosi_debug_level > 3;
                                }

                                # The matched number of attributes equals the number of attributes in urn:*
                                if ($user_ok2 == scalar keys %$key) {
                                    $user_ok++;
                                }
                                AOSICFG::write_ws_log("| $log(cond 10):user_ok:$user_ok;user_ok2:$user_ok2;#urn_keys:" . scalar keys %$key)
                                  if $config::aosi_debug_level > 3;
                            }
                        }
                        if ($user_ok > 0) {
                            AOSICFG::write_ws_log("| $log(attrs): " . AOSICFG::dump_array($attrs)) if $config::aosi_debug_level > 1;
                            $mesg2 = $ldap->add($dn, attrs => $attrs);
                            if ($mesg2->code == 0) {
                                $out_one .= '<status code="0">User added successfully!</status>';
                            }
                            else {
                                $out_one .= '<status code="' . $mesg2->code . '">' . $mesg2->error . '!</status>';
                                $err_num++;
                                $code += $mesg2->code;
                            }
                        }
                        else {
                            $code += 84;
                            $out_one .= '<status code="-84">You can not add this user, due to limitation in your administrator attribute!</status>';
                            $err_num++;
                        }
                    }
                    else {
                        $code += abs(61);
                        $out_one .= '<status code="-61">' . $restriction_msg . '</status>';
                        $err_num++;
                    }
                }
                ($plugin_hash) = Plugins::afterAddUser($ent, $out_one . '</entry>', $user, $pwd);
                ($plugin_panic, $plugin_code, $plugin_xml) = AOSICFG::process_plugin_hash($plugin_hash, 'after');
                $out_one .= $plugin_xml;
            }
            else {
                $code += abs($plugin_code);
                $out_one .= '<status code="' . $code . '">Plugin stopped the action!</status>';
                $err_num++;
            }
            $out_one .= '</entry>';

            #$out .= $out_one;
            $hash->{'result'} .= $out_one;
            AOSICFG::write_action_log($user, $log, $dn, $code, join(",", @req_attrs));
        }
        $hash->{'result'} .= '</ldap>';
        if ($code == 0) {
            $hash->{'code'}  = $code;
            $hash->{'error'} = '';
        }
        else {
            $hash->{'code'}  = $code;
            $hash->{'error'} = $err_num . ' error(s) occured!';
        }
        $mesg2 = $ldap->unbind;
    }
    else {
        $hash->{'code'}  = $auth_code;
        $hash->{'error'} = $auth_error;
        $hash->{'result'} =
            '<?xml version="1.0"?><ldap action="' . $log
          . '" version="'
          . $config::aosi_version
          . '"><entry dn="'
          . $dn
          . '"><status code="'
          . $auth_code . '">'
          . $auth_error
          . '</status></entry></ldap>';
    }

    AOSICFG::write_ws_log("\\ $log(out):" . AOSICFG::dump_hash($hash)) if $config::aosi_debug_level > 0;

    #return $hash;
    if ($type > 0) {
        return $hash;
    }
    else {
        return hash2data($hash, { 'code' => 'int' });
    }
}

######################################################################
#
######################################################################
sub ldapModifyAttribute {
    my ($service, $user, $pwd, $base, $xml_in) = @_;
    my ($hash, $log);
    $log = "ldapModifyAttribute";

    AOSICFG::write_ws_log("/ $log(in):user:$user;base:$base;xml:$xml_in") if $config::aosi_debug_level > 0;
    AOSICFG::write_ws_log("| $log: IP:$AOSI::client_ip;")                 if $config::aosi_debug_level > 1;
    $hash = AOSICFG::ldapGenModify($log, $user, $pwd, $base, $xml_in, 2, 0);
    AOSICFG::write_ws_log("\\ $log(out):" . AOSICFG::dump_hash($hash)) if $config::aosi_debug_level > 0;

    #return $hash;
    return hash2data($hash, { 'code' => 'int' });
}

######################################################################
#
######################################################################
sub userModifyAttribute {
    my ($service, $user, $pwd, $base, $xml_in) = @_;
    my ($hash, $log);
    $log = "userModifyAttribute";

    AOSICFG::write_ws_log("/ $log(in):user:$user;base:$base;xml:$xml_in") if $config::aosi_debug_level > 0;
    AOSICFG::write_ws_log("| $log: IP:$AOSI::client_ip;")                 if $config::aosi_debug_level > 1;
    $hash = AOSICFG::ldapGenModify($log, $user, $pwd, $base, $xml_in, 2, 1);
    AOSICFG::write_ws_log("\\ $log(out):" . AOSICFG::dump_hash($hash)) if $config::aosi_debug_level > 0;

    #return $hash;
    return hash2data($hash, { 'code' => 'int' });
}

######################################################################
#
######################################################################
sub ldapAddAttribute {
    my ($service, $user, $pwd, $base, $xml_in) = @_;
    my ($hash, $log);
    $log = "ldapAddAttribute";

    AOSICFG::write_ws_log("/ $log(in):user:$user;base:$base;xml:$xml_in") if $config::aosi_debug_level > 0;
    AOSICFG::write_ws_log("| $log: IP:$AOSI::client_ip;")                 if $config::aosi_debug_level > 1;
    $hash = AOSICFG::ldapGenModify($log, $user, $pwd, $base, $xml_in, 1, 0);
    AOSICFG::write_ws_log("\\ $log(out):" . AOSICFG::dump_hash($hash)) if $config::aosi_debug_level > 0;

    #return $hash;
    return hash2data($hash, { 'code' => 'int' });
}

######################################################################
#
######################################################################
sub userAddAttribute {
    my ($service, $user, $pwd, $base, $xml_in) = @_;
    my ($hash, $log);
    $log = "userAddAttribute";

    AOSICFG::write_ws_log("/ $log(in):user:$user;base:$base;xml:$xml_in") if $config::aosi_debug_level > 0;
    AOSICFG::write_ws_log("| $log: IP:$AOSI::client_ip;")                 if $config::aosi_debug_level > 1;
    $hash = AOSICFG::ldapGenModify($log, $user, $pwd, $base, $xml_in, 1, 1);
    AOSICFG::write_ws_log("\\ $log(out):" . AOSICFG::dump_hash($hash)) if $config::aosi_debug_level > 0;

    #return $hash;
    return hash2data($hash, { 'code' => 'int' });
}

######################################################################
#
######################################################################
sub ldapDeleteAttribute {
    my ($service, $user, $pwd, $base, $xml_in) = @_;
    my ($hash, $log);
    $log = "ldapDeleteAttribute";

    AOSICFG::write_ws_log("/ $log(in):user:$user;base:$base;xml:$xml_in") if $config::aosi_debug_level > 0;
    AOSICFG::write_ws_log("| $log: IP:$AOSI::client_ip;")                 if $config::aosi_debug_level > 1;
    $hash = AOSICFG::ldapGenModify($log, $user, $pwd, $base, $xml_in, 3, 0);
    AOSICFG::write_ws_log("\\ $log(out):" . AOSICFG::dump_hash($hash)) if $config::aosi_debug_level > 0;

    #return $hash;
    return hash2data($hash, { 'code' => 'int' });
}

######################################################################
#
######################################################################
sub userDeleteAttribute {
    my ($service, $user, $pwd, $base, $xml_in) = @_;
    my ($hash, $log);
    $log = "userDeleteAttribute";

    AOSICFG::write_ws_log("/ $log(in):user:$user;base:$base;xml:$xml_in") if $config::aosi_debug_level > 0;
    AOSICFG::write_ws_log("| $log: IP:$AOSI::client_ip;")                 if $config::aosi_debug_level > 1;
    $hash = AOSICFG::ldapGenModify($log, $user, $pwd, $base, $xml_in, 3, 1);
    AOSICFG::write_ws_log("\\ $log(out):" . AOSICFG::dump_hash($hash)) if $config::aosi_debug_level > 0;

    #return $hash;
    return hash2data($hash, { 'code' => 'int' });
}

######################################################################
#
######################################################################
sub ldapUserExists {
    my ($service, $user, $pwd, $base, $uid, $type) = @_;
    my ($hash, $log);
    $log = "ldapUserExists";

    AOSICFG::write_ws_log("/ $log(in):user:$user;base:$base;uid:$uid;type:$type") if $config::aosi_debug_level > 0;
    AOSICFG::write_ws_log("| $log: IP:$AOSI::client_ip;")                         if $config::aosi_debug_level > 1;

    $hash = AOSICFG::ldapGenSearch($log, $user, $pwd, $base, '&(uid=' . $uid . ')', 'uid', 0, -1, 0);
    AOSICFG::write_ws_log("\\ $log(out):" . AOSICFG::dump_hash($hash)) if $config::aosi_debug_level > 0;
    if ($type > 0) {
        return $hash;
    }
    else {
        return hash2data($hash, { 'code' => 'int' });
    }
}

######################################################################
#
######################################################################
sub ldapBinSearch {
    my ($service, $user, $pwd, $base, $filter, $attribute, $md5, $type) = @_;
    my ($hash, $log);
    $log = "ldapBinSearch";

    AOSICFG::write_ws_log("/ $log(in):user:$user;base:$base;filter:$filter;attrib:$attribute;md5:$md5;type:$type") if $config::aosi_debug_level > 0;
    AOSICFG::write_ws_log("| $log: IP:$AOSI::client_ip;")                                                          if $config::aosi_debug_level > 1;

    if (AOSICFG::is_bin($attribute)) {
        $hash = AOSICFG::ldapGenSearch($log, $user, $pwd, $base, $filter, $attribute, 0, -1, 0, 1, $md5);
    }
    else {
        $hash->{'error'} = $attribute . ' is not binary!';
        $hash->{'code'}  = -33;
    }
    AOSICFG::write_ws_log("\\ $log(out):" . AOSICFG::dump_hash($hash, 'result')) if $config::aosi_debug_level > 0;
    if ($type > 0) {
        return $hash;
    }
    else {
        return hash2data($hash, { 'code' => 'int' });
    }
}

######################################################################
#
######################################################################
sub ldapSearch {
    my ($service, $user, $pwd, $base, $filter, $attributes, $type) = @_;
    my ($hash, $log);
    $log = "ldapSearch";

    AOSICFG::write_ws_log("/ $log(in):user:$user;base:$base;filter:$filter;attrib:$attributes;type:$type") if $config::aosi_debug_level > 0;
    AOSICFG::write_ws_log("| $log: IP:$AOSI::client_ip;")                                                  if $config::aosi_debug_level > 1;
    $hash = AOSICFG::ldapGenSearch($log, $user, $pwd, $base, $filter, $attributes, 0, -1, 0);
    AOSICFG::write_ws_log("\\ $log(out):" . AOSICFG::dump_hash($hash)) if $config::aosi_debug_level > 0;
    if ($type > 0) {
        return $hash;
    }
    else {
        return hash2data($hash, { 'code' => 'int' });
    }
}

######################################################################
#
######################################################################
sub ldapList {
    my ($service, $user, $pwd, $base, $filter, $attributes, $from, $size) = @_;
    my ($hash,  $log,   $ldap,    $hostname, $port,     $auth_code, $auth_error, $is_admin, $dummy);
    my ($mesg,  $mesg2, $mesg3,   %options,  %options2, $code,      $to,         $i,        $entry, $entry2, $dn, $uid, $cur_page);
    my (@attrs, $out,   $my_size, $attr,     $temp,     $tld,       @entries,    $cur_attr, $val, @values, $my_filter, $admin_filter, $ldap_error);
    $log = "ldapList";

    AOSICFG::write_ws_log("/ $log(in):user:$user;base:$base;filter:$filter;attrib:$attributes;from:$from;size:$size;") if $config::aosi_debug_level > 0;
    AOSICFG::write_ws_log("| $log: IP:$AOSI::client_ip;")                                                              if $config::aosi_debug_level > 1;

    ($hostname, $port) = AOSICFG::parse_server($config::ldap_server);
    ($ldap, $ldap_error) = AOSICFG::ldap_init($hostname, $port);
    $dn = AOSICFG::get_dn($user, $base);
    ($auth_code, $auth_error, $is_admin, $admin_filter) = AOSICFG::ck_auth($ldap, AOSICFG::get_dn($user, $base), $pwd, 1, $ldap_error);
    Encode::_utf8_on($admin_filter);
    if ($is_admin == 0 and $ldap) {
        if ($auth_code == 0) {
            $auth_code  = -15;
            $auth_error = "User $user does not have admin privileges!";
        }
        if ($from > -1) {
            $hash->{'from'}  = 0;
            $hash->{'size'}  = 0;
            $hash->{'count'} = 0;
        }
    }
    if ($auth_code == 0) {
        $mesg = $ldap->bind($config::ldap_user, password => AOSICFG::get_pwd($config::ldap_pwd_file), version => 3);
        if ($mesg->code != 0) {
            AOSICFG::write_ws_log("| !!!$log: invalid admin password in " . $config::ldap_pwd_file);
        }
        AOSICFG::write_ws_log("| $log(bind_as_admin): " . $mesg->code) if $config::aosi_debug_level > 2;
        if ($filter =~ /^\(.*\)$/) {
            $my_filter = '(&' . $filter . '' . $admin_filter . ')';
        }
        else {
            $my_filter = '(&(' . $filter . ')' . $admin_filter . ')';
        }
        AOSICFG::write_ws_log("| $log(my_filter): " . $my_filter) if $config::aosi_debug_level > 2;
        %options = (
            base      => $base,
            filter    => $my_filter,
            attrs     => ['dn'],
            timelimit => 180
        );
        my $t1 = time;
        AOSICFG::write_ws_log("| $log:time1= " . $t1) if $config::aosi_debug_level > 2;
        $mesg2 = $ldap->search(%options);
        my $t2 = time;
        AOSICFG::write_ws_log("| $log:time2= " . $t2) if $config::aosi_debug_level > 2;
        AOSICFG::write_ws_log("| $log:time2-time1= " . sprintf("%d", $t2 - $t1)) if $config::aosi_debug_level > 2;
        $code           = $mesg2->code;
        $hash->{'code'} = $code;
        $my_size        = $size > 100 ? 100 : $size;
        $to             = $from + $my_size - 1;
        AOSICFG::write_ws_log("| $log:to=" . $to) if $config::aosi_debug_level > 1;
        $i = -1;
        my $t3 = time;
        AOSICFG::write_ws_log("| $log:time3= " . $t3) if $config::aosi_debug_level > 2;
        AOSICFG::write_ws_log("| $log:time3-time2= " . sprintf("%d", $t3 - $t2)) if $config::aosi_debug_level > 2;
        $hash->{'from'}  = 0;
        $hash->{'size'}  = 0;
        $hash->{'count'} = 0;

        if ($code == 0) {
            @entries = $mesg2->entries;
            foreach $entry (sort bydn @entries) {
                $i++;
                next if ($i < $from);
                last if ($i > $to);
                $dn = $entry->dn();
                $out .= "<entry num=\"$i\" dn=\"$dn\">";
                AOSICFG::write_ws_log("| $log:dn=" . $dn) if $config::aosi_debug_level > 2;
                ($uid, $temp, $tld) = split(/,/, $dn);
                ($temp, $uid) = split(/=/, $uid);
                AOSICFG::write_ws_log("| $log:uid=" . $uid) if $config::aosi_debug_level > 3;
                %options2 = (
                    base      => $base,
                    filter    => '(uid=' . $uid . ')',
                    timelimit => 180
                );

                if ($attributes) {
                    @attrs = split(/[,; ]/, $attributes);
                    $options2{'attrs'} = \@attrs;
                }
                $mesg3  = $ldap->search(%options2);
                $entry2 = ($mesg3->entries)[0];
                if ($entry2) {
                    foreach $attr ($entry2->attributes) {
                        $cur_attr = $attr eq 'homePhone' ? 'homeTelephoneNumber' : $attr;
                        AOSICFG::write_ws_log("| $log(attr): $attr") if $config::aosi_debug_level > 3;
                        $out .= "<attribute ldapname=\"" . $cur_attr . "\">";
                        @values = @{ $entry2->get_value($attr, asref => 1) };
                        foreach $val (@values) {
                            $out .= "<value>";
                            if (AOSICFG::is_bin($attr)) {
                                $out .= encode_base64($val);
                            }
                            else {
                                $out .= XML::Simple::escape_value($dummy, $val);
                            }
                            $out .= "</value>";
                        }
                        $out .= "</attribute>";
                    }
                }
                $out .= "</entry>";

                #AOSICFG::write_action_log($user, $log, $dn, $auth_code, $attributes);
            }
            my $t4 = time;
            AOSICFG::write_ws_log("| $log:time4= " . $t4) if $config::aosi_debug_level > 2;
            AOSICFG::write_ws_log("| $log:time4-time3= " . sprintf("%d", $t4 - $t3)) if $config::aosi_debug_level > 2;
            AOSICFG::write_ws_log("| $log: XML building finished") if $config::aosi_debug_level > 3;
            if ($from > -1) {
                $hash->{'from'}  = $from;
                $hash->{'size'}  = $size;
                $hash->{'count'} = $#entries + 1;
            }
            $out              = "<?xml version=\"1.0\" encoding=\"utf-8\"?><ldap action=\"$log\" version=\"" . $config::aosi_version . "\">$out</ldap>";
            $hash->{'result'} = $out;
            $hash->{'error'}  = '';
        }
        else {
            $hash->{'error'}  = $mesg2->error;
            $hash->{'result'} = '';
        }
        $mesg = $ldap->unbind;
    }
    else {
        $hash->{'count'} = 0;
        $hash->{'size'}  = 0;
        $hash->{'from'}  = 0;
        $hash->{'code'}  = $auth_code;
        $hash->{'error'} = $auth_error;
        $hash->{'result'} =
            '<?xml version="1.0"?><ldap action="' . $log
          . '" version="'
          . $config::aosi_version
          . '"><entry dn="'
          . $dn
          . '"><status code="'
          . $auth_code . '">'
          . $auth_error
          . '</status></entry></ldap>';
    }

    #$hash = AOSICFG::ldapGenSearch($log, $user, $pwd, $base, $filter, $attributes, 0, $from, $size);
    AOSICFG::write_ws_log("\\ $log(out):" . AOSICFG::dump_hash($hash)) if $config::aosi_debug_level > 0;

    #return $hash;
    return hash2data($hash, { 'code' => 'int', 'count' => 'int', 'from' => 'int', 'size' => 'int' });
}

sub bydn {
    $a->dn() cmp $b->dn();
}

######################################################################
#
######################################################################
sub userBinSearch {
    my ($service, $user, $pwd, $base, $filter, $attribute, $md5, $type) = @_;
    my ($hash, $log);
    $log = "userBinSearch";

    AOSICFG::write_ws_log("/ $log(in):user:$user;base:$base;filter:$filter;attrib:$attribute;md5:$md5;type:$type;") if $config::aosi_debug_level > 0;
    AOSICFG::write_ws_log("| $log: IP:$AOSI::client_ip;")                                                           if $config::aosi_debug_level > 1;

    if (AOSICFG::is_bin($attribute)) {
        $hash = AOSICFG::ldapGenSearch($log, $user, $pwd, $base, $filter, $attribute, 1, -1, 0, 1, $md5);
    }
    else {
        $hash->{'code'}   = -34;
        $hash->{'error'}  = $attribute . ' is not binary!';
        $hash->{'result'} = '';
    }
    AOSICFG::write_ws_log("\\ $log(out):" . AOSICFG::dump_hash($hash, 'result')) if $config::aosi_debug_level > 0;
    if ($type > 0) {
        return $hash;
    }
    else {
        return hash2data($hash, { 'code' => 'int' });
    }
}

######################################################################
#
######################################################################
sub anonBinSearch {
    my ($service, $base, $filter, $attribute, $md5, $type) = @_;
    my ($hash, $log, $user, $pwd);
    $log = "anonBinSearch";

    AOSICFG::write_ws_log("/ $log(in):filter:$filter;attrib:$attribute;md5:$md5;type:$type;") if $config::aosi_debug_level > 0;
    AOSICFG::write_ws_log("| $log: IP:$AOSI::client_ip;")                                     if $config::aosi_debug_level > 1;

    if (AOSICFG::is_bin($attribute)) {
        $hash = AOSICFG::ldapGenSearch($log, $user, $pwd, $base, $filter, $attribute, -1, -1, 0, 1, $md5);
    }
    else {
        $hash->{'code'}   = -35;
        $hash->{'error'}  = $attribute . ' is not binary!';
        $hash->{'result'} = '';
    }
    AOSICFG::write_ws_log("\\ $log(out):" . AOSICFG::dump_hash($hash, 'result')) if $config::aosi_debug_level > 0;
    if ($type > 0) {
        return $hash;
    }
    else {
        return hash2data($hash, { 'code' => 'int' });
    }
}

######################################################################
#
######################################################################
sub userSearch {
    my ($service, $user, $pwd, $base, $filter, $attributes, $type) = @_;
    my ($hash, $log);
    $log = "userSearch";

    AOSICFG::write_ws_log("/ $log(in):user:$user;base:$base;filter:$filter;attrib:$attributes;type:$type;") if $config::aosi_debug_level > 0;
    AOSICFG::write_ws_log("| $log: IP:$AOSI::client_ip;")                                                   if $config::aosi_debug_level > 1;
    $hash = AOSICFG::ldapGenSearch($log, $user, $pwd, $base, $filter, $attributes, 1, -1, 0);
    AOSICFG::write_ws_log("\\ $log(out):" . AOSICFG::dump_hash($hash)) if $config::aosi_debug_level > 0;
    if ($type > 0) {
        return $hash;
    }
    else {
        return hash2data($hash, { 'code' => 'int' });
    }
}

######################################################################
#
######################################################################
sub anonSearch {
    my ($service, $base, $filter, $attributes, $type) = @_;
    my ($hash, $log, $user, $pwd);
    $log = "anonSearch";

    AOSICFG::write_ws_log("/ $log(in):filter:$filter;attrib:$attributes;type:$type;") if $config::aosi_debug_level > 0;
    AOSICFG::write_ws_log("| $log: IP:$AOSI::client_ip;")                             if $config::aosi_debug_level > 1;
    $hash = AOSICFG::ldapGenSearch($log, $user, $pwd, $base, $filter, $attributes, -1, -1, 0);
    AOSICFG::write_ws_log("\\ $log(out):" . AOSICFG::dump_hash($hash)) if $config::aosi_debug_level > 0;
    if ($type > 0) {
        return $hash;
    }
    else {
        return hash2data($hash, { 'code' => 'int' });
    }
}

######################################################################
#
######################################################################
sub hash2data {
    my ($hash, $types) = @_;
    my ($key, $type, @out);

    foreach $key (sort keys %$hash) {

        #$type = $types->{$key} || 'string';
        $type = $types->{$key} || 'base64';
        push(@out, SOAP::Data->name($key)->type($type)->value($hash->{$key}));
    }
    return @out;
}

1;
