#!/usr/bin/perl -Tw

BEGIN {
    unshift(@INC, '/usr/lib/aosi');
    unshift(@INC, '/etc/aosi');
}

use config;

#use SOAP::Lite +trace => [qw(debug)];
use SOAP::Lite;
use SOAP::Transport::HTTP;
use AOSI;
use POSIX qw(setsid sigprocmask);
use PreforkSOAP;

$arg = $ARGV[0] || '';

$daemonize       = $arg eq '-d' ? 1 : 0;
$AOSI::client_ip = '';
$log             = '(from command line)';

if ($daemonize == 1) {

    # Be a nice daemon and chdir() to the root so we don't block any unmount attempts
    chdir '/' or die "Can't chdir to /: $!\n";

    # Redirect all warnings to log
    $SIG{__WARN__} = sub { AOSICFG::write_ws_log($_[0]); };

    # Redirect in and out to the bit bucket
    open STDIN,  "</dev/null" or die "Can't read from /dev/null: $!\n";
    open STDOUT, ">/dev/null" or die "Can't write to /dev/null: $!\n";

    # Here we go...
    defined(my $pid = fork) or die "Can't fork: $!\n";
    exit if $pid;
    setsid or die "Can't start new session: $!\n";

    # Now we can redirect the errors, too.
    open STDERR, '>&STDOUT' or die "Can't duplicate stdout: $!\n";
    $log = '(daemonized)';
}

# don't want to die on 'Broken pipe' or Ctrl-C
#$SIG{PIPE} = $SIG{INT} = 'IGNORE';
$SIG{PIPE} = 'IGNORE';
$SIG{CHLD} = 'IGNORE';

my %args;
$args{'LocalPort'}     = $config::ws_port;
$args{'ReuseAddr'}     = 1;
$args{'Listen'}        = 5;
$args{'SSL_key_file'}  = $config::etc_dir . '/certs/aosi_key.pem';
$args{'SSL_cert_file'} = $config::etc_dir . '/certs/aosi_cert.pem';

$daemon = SOAP::Transport::HTTP::Daemon::PreforkSOAP
->new(%args)
->dispatch_to($config::lib_dir, 'AOSI');

if ($SOAP::Lite::VERSION > 0.60) {
    my $ser = $daemon->serializer;
    $ser->ns('urn:AOSI');
}

AOSICFG::write_ws_log("===== Starting server (" . $SOAP::Lite::VERSION . ") " . $daemon->url . " " . $log . " =====");
AOSICFG::write_pid('/var/run/aosi/aosi.pid');

$daemon->handle(\$AOSI::client_ip, 10);

